#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq

from piquasso.api.result import Result


def test_Result_repr(FakeState):
    result = Result(
        state=FakeState(d=3, connector=pq.NumpyConnector()), samples=[1, 2, 3]
    )

    assert (
        repr(result)
        == "Result(samples=[1, 2, 3], state=FakeState(d=3, config=Config(), connector=NumpyConnector()))"  # noqa: E501
    )
